@@ -2,22 +2,23 @@
  * mpfr.c - routines for arbitrary-precision number support in gawk.
  */
 
-/* 
- * Copyright (C) 2012, 2013, 2015 the Free Software Foundation, Inc.
- * 
+/*
+ * Copyright (C) 2012, 2013, 2015, 2017, 2018, 2019, 2021, 2022,
+ * the Free Software Foundation, Inc.
+ *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
- * 
+ *
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
- * 
+ *
  * GAWK is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
@@ -27,6 +28,8 @@
 
 #ifdef HAVE_MPFR
 
+int MPFR_round_mode = 'N';	// default value
+
 #if !defined(MPFR_VERSION_MAJOR) || MPFR_VERSION_MAJOR < 3
 typedef mp_exp_t mpfr_exp_t;
 #endif
@@ -103,28 +106,16 @@ cleanup_mpfr(void)
 /* mpg_node --- allocate a node to store MPFR float or GMP integer */
 
 NODE *
-mpg_node(unsigned int tp)
+mpg_node(unsigned int flags)
 {
-	NODE *r;
-	getnode(r);
-	r->type = Node_val;
+	NODE *r = make_number_node(flags);
 
-	if (tp == MPFN) {
+	if (flags == MPFN)
 		/* Initialize, set precision to the default precision, and value to NaN */
 		mpfr_init(r->mpg_numbr);
-		r->flags = MPFN;
-	} else {
+	else
 		/* Initialize and set value to 0 */
 		mpz_init(r->mpg_i);
-		r->flags = MPZN;
-	}
-	
-	r->valref = 1;
-	r->flags |= MALLOC|NUMBER|NUMCUR;
-	r->stptr = NULL;
-	r->stlen = 0;
-	r->wstptr = NULL;
-	r->wstlen = 0;
 	return r;
 }
 
@@ -151,7 +142,7 @@ mpg_make_number(double x)
 	return r;
 }
 
-/* mpg_strtoui --- assign arbitrary-precision integral value from a string */ 
+/* mpg_strtoui --- assign arbitrary-precision integral value from a string */
 
 int
 mpg_strtoui(mpz_ptr zi, char *str, size_t len, char **end, int base)
@@ -185,7 +176,7 @@ mpg_strtoui(mpz_ptr zi, char *str, size_t len, char **end, int base)
 		case '8':
 		case '9':
 			if (base == 8)
-				goto done;
+				base = 10;
 			break;
 		case 'a':
 		case 'b':
@@ -256,7 +247,7 @@ mpg_maybe_float(const char *str, int use_locale)
 
 /* mpg_zero --- initialize with arbitrary-precision integer(GMP) and set value to zero */
 
-static inline void
+void
 mpg_zero(NODE *n)
 {
 	if (is_mpg_float(n)) {
@@ -264,7 +255,7 @@ mpg_zero(NODE *n)
 		n->flags &= ~MPFN;
 	}
 	if (! is_mpg_integer(n)) {
-		mpz_init(n->mpg_i);	/* this also sets its value to 0 */ 
+		mpz_init(n->mpg_i);	/* this also sets its value to 0 */
 		n->flags |= MPZN;
 	} else
 		mpz_set_si(n->mpg_i, 0);
@@ -280,7 +271,7 @@ force_mpnum(NODE *n, int do_nondec, int use_locale)
 	char save;
 	int tval, base = 10;
 
-	if (n->stlen == 0) {
+	if (n->stlen == 0 || (n->flags & REGEX) != 0) {
 		mpg_zero(n);
 		return false;
 	}
@@ -302,10 +293,21 @@ force_mpnum(NODE *n, int do_nondec, int use_locale)
 	else
 		cp1 = cp;
 
-	if (do_nondec)
-		base = get_numbase(cp1, use_locale);
+	/*
+	 * Maybe "+" or "-" was the field.  mpg_strtoui
+	 * won't check for that and set errno, so we have
+	 * to check manually.
+	 */
+	if (*cp1 == '\0') {
+		*cpend = save;
+		mpg_zero(n);
+		return false;
+	}
 
-	if (! mpg_maybe_float(cp1, use_locale)) {
+	if (do_nondec)
+		base = get_numbase(cp1, cpend - cp1, use_locale);
+
+	if (base != 10 || ! mpg_maybe_float(cp1, use_locale)) {
 		mpg_zero(n);
 		errno = 0;
 		mpg_strtoui(n->mpg_i, cp1, cpend - cp1, & ptr, base);
@@ -326,16 +328,18 @@ force_mpnum(NODE *n, int do_nondec, int use_locale)
 
 	errno = 0;
 	tval = mpfr_strtofr(n->mpg_numbr, cp, & ptr, base, ROUND_MODE);
+	if (mpfr_nan_p(n->mpg_numbr) && *cp == '-')
+		tval = mpfr_setsign(n->mpg_numbr, n->mpg_numbr, 1, ROUND_MODE);
 	IEEE_FMT(n->mpg_numbr, tval);
 done:
 	/* trailing space is OK for NUMBER */
-	while (isspace((unsigned char) *ptr))
+	while (ptr < cpend && isspace((unsigned char) *ptr))
 		ptr++;
 	*cpend = save;
 	if (errno == 0 && ptr == cpend)
 		return true;
 	errno = 0;
-	return false; 
+	return false;
 }
 
 /* mpg_force_number --- force a value to be a multiple-precision number */
@@ -343,20 +347,68 @@ done:
 static NODE *
 mpg_force_number(NODE *n)
 {
-	unsigned int newflags = 0;
+	char *cp, *cpend;
+
+	if (n->type == Node_elem_new) {
+		n->type = Node_val;
+		n->flags &= ~STRING;
+		n->stptr[0] = '0';	// STRCUR is still set
+		n->stlen = 1;
 
-	if (is_mpg_number(n) && (n->flags & NUMCUR) != 0)
 		return n;
-
-	if ((n->flags & MAYBE_NUM) != 0) {
-		n->flags &= ~MAYBE_NUM;
-		newflags = NUMBER;
 	}
 
+	if ((n->flags & NUMCUR) != 0)
+		return n;
+	n->flags |= NUMCUR;
+
+	/* Trim leading white space, bailing out if there's nothing else */
+	for (cp = n->stptr, cpend = cp + n->stlen;
+	     cp < cpend && isspace((unsigned char) *cp); cp++)
+		continue;
+
+	if (cp == cpend)
+		goto badnum;
+
+	/* At this point, we know the string is not entirely white space */
+	/* Trim trailing white space */
+	while (isspace((unsigned char) cpend[-1]))
+		cpend--;
+
+	/*
+	 * 2/2007:
+	 * POSIX, by way of severe language lawyering, seems to
+	 * allow things like "inf" and "nan" to mean something.
+	 * So if do_posix, the user gets what he deserves.
+	 * This also allows hexadecimal floating point. Ugh.
+	 */
+	if (! do_posix) {
+		if (is_alpha((unsigned char) *cp))
+			goto badnum;
+		else if (is_ieee_magic_val(cp)) {
+			if (cpend != cp + 4)
+				goto badnum;
+			/* else
+				fall through */
+		}
+		/* else
+			fall through */
+	}
+	/* else POSIX, so
+		fall through */
+
 	if (force_mpnum(n, (do_non_decimal_data && ! do_traditional), true)) {
-		n->flags |= newflags;
-		n->flags |= NUMCUR;
-	}
+		if ((n->flags & USER_INPUT) != 0) {
+			/* leave USER_INPUT set to indicate a strnum */
+			n->flags &= ~STRING;
+			n->flags |= NUMBER;
+		}
+	} else
+		n->flags &= ~USER_INPUT;
+	return n;
+badnum:
+	mpg_zero(n);
+	n->flags &= ~USER_INPUT;
 	return n;
 }
 
@@ -368,6 +420,11 @@ mpg_format_val(const char *format, int index, NODE *s)
 	NODE *dummy[2], *r;
 	unsigned int oflags;
 
+	if (out_of_range(s)) {
+		const char *result = format_nan_inf(s, 'g');
+		return make_string(result, strlen(result));
+	}
+
 	/* create dummy node for a sole use of format_tree */
 	dummy[1] = s;
 	oflags = s->flags;
@@ -375,7 +432,7 @@ mpg_format_val(const char *format, int index, NODE *s)
 	if (is_mpg_integer(s) || mpfr_integer_p(s->mpg_numbr)) {
 		/* integral value, use %d */
 		r = format_tree("%d", 2, dummy, 2);
-		s->stfmt = -1;
+		s->stfmt = STFMT_UNUSED;
 	} else {
 		r = format_tree(format, fmt_list[index]->stlen, dummy, 2);
 		assert(r != NULL);
@@ -383,12 +440,12 @@ mpg_format_val(const char *format, int index, NODE *s)
 	}
 	s->flags = oflags;
 	s->stlen = r->stlen;
-	if ((s->flags & STRCUR) != 0)
+	if ((s->flags & (MALLOC|STRCUR)) == (MALLOC|STRCUR))
 		efree(s->stptr);
 	s->stptr = r->stptr;
-	freenode(r);	/* Do not unref(r)! We want to keep s->stptr == r->stpr.  */
- 
 	s->flags |= STRCUR;
+	s->strndmode = MPFR_round_mode;
+	freenode(r);	/* Do not unref(r)! We want to keep s->stptr == r->stpr.  */
 	free_wstr(s);
 	return s;
 }
@@ -428,10 +485,57 @@ mpg_cmp(const NODE *t1, const NODE *t2)
 	return cmp_awknums(t1, t2);
 }
 
+/* mpg_cmp_as_numbers --- compare two numbers, similar to doubles */
+
+bool
+mpg_cmp_as_numbers(const NODE *t1, const NODE *t2, scalar_cmp_t comparison_type)
+{
+	/*
+	 * This routine provides numeric comparisons that should work
+	 * the same as in C.  It should NOT be used for sorting.
+	 */
+
+	bool t1_nan = mpfr_nan_p(t1->mpg_numbr);
+	bool t2_nan = mpfr_nan_p(t2->mpg_numbr);
+	bool ret = false;
+
+	// MPFR is different than native doubles...
+	if (t1_nan || t2_nan)
+		return comparison_type == SCALAR_NEQ;
+
+	int di = mpg_cmp(t1, t2);
+
+	switch (comparison_type) {
+	case SCALAR_EQ:
+		ret = (di == 0);
+		break;
+	case SCALAR_NEQ:
+		ret = (di != 0);
+		break;
+	case SCALAR_LT:
+		ret = (di < 0);
+		break;
+	case SCALAR_LE:
+		ret = (di <= 0);
+		break;
+	case SCALAR_GT:
+		ret = (di > 0);
+		break;
+	case SCALAR_GE:
+		ret = (di >= 0);
+		break;
+	default:
+		cant_happen("invalid comparison type %d", (int) comparison_type);
+		break;
+	}
+
+	return ret;
+}
+
 
 /*
- * mpg_update_var --- update NR or FNR. 
- *	NR_node->var_value(mpz_t) = MNR(mpz_t) * LONG_MAX + NR(long) 
+ * mpg_update_var --- update NR or FNR.
+ *	NR_node->var_value(mpz_t) = MNR(mpz_t) * LONG_MAX + NR(long)
  */
 
 NODE *
@@ -448,7 +552,7 @@ mpg_update_var(NODE *n)
 		nr = FNR;
 		nq = MFNR;
 	} else
-		cant_happen();
+		cant_happen("invalid node for mpg_update_var%s", "");
 
 	if (mpz_sgn(nq) == 0) {
 		/* Efficiency hack similar to that for AWKNUM */
@@ -480,12 +584,12 @@ mpg_set_var(NODE *n)
 	else if (n == FNR_node)
 		nq = MFNR;
 	else
-		cant_happen();
+		cant_happen("invalid node for mpg_set_var%s", "");
 
 	if (is_mpg_integer(val))
 		r = val->mpg_i;
 	else {
-		/* convert float to integer */ 
+		/* convert float to integer */
 		mpfr_get_z(mpzval, val->mpg_numbr, MPFR_RNDZ);
 		r = mpzval;
 	}
@@ -506,11 +610,11 @@ set_PREC()
 		mpfr_exp_t emax;
 		mpfr_exp_t emin;
 	} ieee_fmts[] = {
-{ "half",	11,	16,	-23	},	/* binary16 */
-{ "single",	24,	128,	-148	},	/* binary32 */
-{ "double",	53,	1024,	-1073	},	/* binary64 */
-{ "quad",	113,	16384,	-16493	},	/* binary128 */
-{ "oct",	237,	262144,	-262377	},	/* binary256, not in the IEEE 754-2008 standard */
+		{ "half",	11,	16,	-23	},	/* binary16 */
+		{ "single",	24,	128,	-148	},	/* binary32 */
+		{ "double",	53,	1024,	-1073	},	/* binary64 */
+		{ "quad",	113,	16384,	-16493	},	/* binary128 */
+		{ "oct",	237,	262144,	-262377	},	/* binary256, not in the IEEE 754-2008 standard */
 
 		/*
  		 * For any bitwidth = 32 * k ( k >= 4),
@@ -523,11 +627,9 @@ set_PREC()
 	if (! do_mpfr)
 		return;
 
-	val = PREC_node->var_value;
-	if ((val->flags & MAYBE_NUM) != 0)
-		force_number(val);
+	val = fixtype(PREC_node->var_value);
 
-	if ((val->flags & STRCUR) != 0) {
+	if ((val->flags & STRING) != 0) {
 		int i, j;
 
 		/* emulate IEEE-754 binary format */
@@ -553,7 +655,7 @@ set_PREC()
 
 	if (prec <= 0) {
 		force_number(val);
-		prec = get_number_si(val);		
+		prec = get_number_si(val);
 		if (prec < MPFR_PREC_MIN || prec > MPFR_PREC_MAX) {
 			force_string(val);
 			warning(_("PREC value `%.*s' is invalid"), (int) val->stlen, val->stptr);
@@ -588,12 +690,12 @@ get_rnd_mode(const char rmode)
 #if defined(MPFR_VERSION_MAJOR) && MPFR_VERSION_MAJOR > 2
 	case 'A':
 	case 'a':
-		return MPFR_RNDA;	/* round away from zero (IEEE-754 roundTiesToAway) */
+		return MPFR_RNDA;	/* round away from zero */
 #endif
 	default:
 		break;
 	}
-	return -1;
+	return (mpfr_rnd_t) -1;
 }
 
 /*
@@ -605,7 +707,7 @@ void
 set_ROUNDMODE()
 {
 	if (do_mpfr) {
-		mpfr_rnd_t rndm = -1;
+		mpfr_rnd_t rndm = (mpfr_rnd_t) -1;
 		NODE *n;
 		n = force_string(ROUNDMODE_node->var_value);
 		if (n->stlen == 1)
@@ -613,8 +715,9 @@ set_ROUNDMODE()
 		if (rndm != -1) {
 			mpfr_set_default_rounding_mode(rndm);
 			ROUND_MODE = rndm;
+			MPFR_round_mode = n->stptr[0];
 		} else
-			warning(_("RNDMODE value `%.*s' is invalid"), (int) n->stlen, n->stptr);
+			warning(_("ROUNDMODE value `%.*s' is invalid"), (int) n->stlen, n->stptr);
 	}
 }
 
@@ -673,13 +776,15 @@ do_mpfr_atan2(int nargs)
 	mpfr_ptr p1, p2;
 	int tval;
 
+	check_exact_args(nargs, "atan2", 2);
+
 	t2 = POP_SCALAR();
 	t1 = POP_SCALAR();
 
 	if (do_lint) {
-		if ((t1->flags & (NUMCUR|NUMBER)) == 0)
+		if ((fixtype(t1)->flags & NUMBER) == 0)
 			lintwarn(_("atan2: received non-numeric first argument"));
-		if ((t2->flags & (NUMCUR|NUMBER)) == 0)
+		if ((fixtype(t2)->flags & NUMBER) == 0)
 			lintwarn(_("atan2: received non-numeric second argument"));
 	}
 	force_number(t1);
@@ -688,7 +793,7 @@ do_mpfr_atan2(int nargs)
 	p1 = MP_FLOAT(t1);
 	p2 = MP_FLOAT(t2);
 	res = mpg_float();
-	/* See MPFR documentation for handling of special values like +inf as an argument */ 
+	/* See MPFR documentation for handling of special values like +inf as an argument */
 	tval = mpfr_atan2(res->mpg_numbr, p1, p2, ROUND_MODE);
 	IEEE_FMT(res->mpg_numbr, tval);
 
@@ -701,20 +806,26 @@ do_mpfr_atan2(int nargs)
 
 static inline NODE *
 do_mpfr_func(const char *name,
-		int (*mpfr_func)(),	/* putting argument types just gets the compiler confused */
-		int nargs)
+		int (*mpfr_func)(mpfr_ptr, mpfr_srcptr, mpfr_rnd_t),
+		int nargs, bool warn_negative)
 {
 	NODE *t1, *res;
 	mpfr_ptr p1;
 	int tval;
 	mpfr_prec_t argprec;
 
+	check_exact_args(nargs, name, 1);
+
 	t1 = POP_SCALAR();
-	if (do_lint && (t1->flags & (NUMCUR|NUMBER)) == 0)
+	if (do_lint && (fixtype(t1)->flags & NUMBER) == 0)
 		lintwarn(_("%s: received non-numeric argument"), name);
 
 	force_number(t1);
 	p1 = MP_FLOAT(t1);
+	if (warn_negative && mpfr_sgn(p1) < 0) {
+		force_string(t1);
+		warning(_("%s: received negative argument %.*s"), name, (int) t1->stlen, t1->stptr);
+	}
 	res = mpg_float();
 	if ((argprec = mpfr_get_prec(p1)) > default_prec)
 		mpfr_set_prec(res->mpg_numbr, argprec);	/* needed at least for sqrt() */
@@ -724,9 +835,9 @@ do_mpfr_func(const char *name,
 	return res;
 }
 
-#define SPEC_MATH(X)				\
+#define SPEC_MATH(X, WN)				\
 NODE *result;					\
-result = do_mpfr_func(#X, mpfr_##X, nargs);	\
+result = do_mpfr_func(#X, mpfr_##X, nargs, WN);	\
 return result
 
 /* do_mpfr_sin --- do the sin function */
@@ -734,7 +845,7 @@ return result
 NODE *
 do_mpfr_sin(int nargs)
 {
-	SPEC_MATH(sin);
+	SPEC_MATH(sin, false);
 }
 
 /* do_mpfr_cos --- do the cos function */
@@ -742,7 +853,7 @@ do_mpfr_sin(int nargs)
 NODE *
 do_mpfr_cos(int nargs)
 {
-	SPEC_MATH(cos);
+	SPEC_MATH(cos, false);
 }
 
 /* do_mpfr_exp --- exponential function */
@@ -750,7 +861,7 @@ do_mpfr_cos(int nargs)
 NODE *
 do_mpfr_exp(int nargs)
 {
-	SPEC_MATH(exp);
+	SPEC_MATH(exp, false);
 }
 
 /* do_mpfr_log --- the log function */
@@ -758,7 +869,7 @@ do_mpfr_exp(int nargs)
 NODE *
 do_mpfr_log(int nargs)
 {
-	SPEC_MATH(log);
+	SPEC_MATH(log, true);
 }
 
 /* do_mpfr_sqrt --- do the sqrt function */
@@ -766,7 +877,7 @@ do_mpfr_log(int nargs)
 NODE *
 do_mpfr_sqrt(int nargs)
 {
-	SPEC_MATH(sqrt);
+	SPEC_MATH(sqrt, true);
 }
 
 /* do_mpfr_int --- convert double to int for awk */
@@ -776,8 +887,10 @@ do_mpfr_int(int nargs)
 {
 	NODE *tmp, *r;
 
+	check_exact_args(nargs, "int", 1);
+
 	tmp = POP_SCALAR();
-	if (do_lint && (tmp->flags & (NUMCUR|NUMBER)) == 0)
+	if (do_lint && (fixtype(tmp)->flags & NUMBER) == 0)
 		lintwarn(_("int: received non-numeric argument"));
 	force_number(tmp);
 
@@ -806,8 +919,10 @@ do_mpfr_compl(int nargs)
 	NODE *tmp, *r;
 	mpz_ptr zptr;
 
+	check_exact_args(nargs, "compl", 1);
+
 	tmp = POP_SCALAR();
-	if (do_lint && (tmp->flags & (NUMCUR|NUMBER)) == 0)
+	if (do_lint && (fixtype(tmp)->flags & NUMBER) == 0)
 		lintwarn(_("compl: received non-numeric argument"));
 
 	force_number(tmp);
@@ -818,28 +933,26 @@ do_mpfr_compl(int nargs)
 			/* [+-]inf or NaN */
 			return tmp;
 		}
-		if (do_lint) {
-			if (mpfr_sgn(p) < 0)
-				lintwarn("%s",
-			mpg_fmt(_("compl(%Rg): negative value will give strange results"), p)
+		if (mpfr_sgn(p) < 0)
+			fatal("%s",
+			mpg_fmt(_("compl(%Rg): negative value is not allowed"), p)
 					);
+		if (do_lint) {
 			if (! mpfr_integer_p(p))
 				lintwarn("%s",
 			mpg_fmt(_("comp(%Rg): fractional value will be truncated"), p)
 					);
 		}
-		
+
 		mpfr_get_z(mpzval, p, MPFR_RNDZ);	/* float to integer conversion */
 		zptr = mpzval;
 	} else {
-		/* (tmp->flags & MPZN) != 0 */ 
+		/* (tmp->flags & MPZN) != 0 */
 		zptr = tmp->mpg_i;
-		if (do_lint) {
-			if (mpz_sgn(zptr) < 0)
-				lintwarn("%s",
-			mpg_fmt(_("cmpl(%Zd): negative values will give strange results"), zptr)
+		if (mpz_sgn(zptr) < 0)
+				fatal("%s",
+			mpg_fmt(_("compl(%Zd): negative values are not allowed"), zptr)
 					);
-		}
 	}
 
 	r = mpg_integer();
@@ -855,7 +968,7 @@ get_intval(NODE *t1, int argnum, const char *op)
 {
 	mpz_ptr pz;
 
-	if (do_lint && (t1->flags & (NUMCUR|NUMBER)) == 0)
+	if (do_lint && (fixtype(t1)->flags & NUMBER) == 0)
 		lintwarn(_("%s: received non-numeric argument #%d"), op, argnum);
 
 	(void) force_number(t1);
@@ -875,13 +988,13 @@ get_intval(NODE *t1, int argnum, const char *op)
 			return pz;	/* should be freed */
 		}
 
-		if (do_lint) {
-			if (mpfr_sgn(left) < 0)
-				lintwarn("%s",
-		mpg_fmt(_("%s: argument #%d negative value %Rg will give strange results"),
+		if (mpfr_sgn(left) < 0)
+			fatal("%s",
+		mpg_fmt(_("%s: argument #%d negative value %Rg is not allowed"),
 					op, argnum, left)
 				);
 
+		if (do_lint) {
 			if (! mpfr_integer_p(left))
 				lintwarn("%s",
 		mpg_fmt(_("%s: argument #%d fractional value %Rg will be truncated"),
@@ -893,16 +1006,15 @@ get_intval(NODE *t1, int argnum, const char *op)
 		mpz_init(pz);
 		mpfr_get_z(pz, left, MPFR_RNDZ);	/* float to integer conversion */
 		return pz;	/* should be freed */
-	} 
-	/* (t1->flags & MPZN) != 0 */ 
+	}
+	/* (t1->flags & MPZN) != 0 */
 	pz = t1->mpg_i;
-	if (do_lint) {
-		if (mpz_sgn(pz) < 0)
-			lintwarn("%s",
-	mpg_fmt(_("%s: argument #%d negative value %Zd will give strange results"),
+	if (mpz_sgn(pz) < 0)
+		fatal("%s",
+	mpg_fmt(_("%s: argument #%d negative value %Zd is not allowed"),
 					op, argnum, pz)
 				);
-	}
+
 	return pz;	/* must not be freed */
 }
 
@@ -927,7 +1039,9 @@ do_mpfr_lshift(int nargs)
 	NODE *t1, *t2, *res;
 	unsigned long shift;
 	mpz_ptr pz1, pz2;
- 
+
+	check_exact_args(nargs, "lshift", 2);
+
 	t2 = POP_SCALAR();
 	t1 = POP_SCALAR();
 
@@ -959,7 +1073,9 @@ do_mpfr_rshift(int nargs)
 	NODE *t1, *t2, *res;
 	unsigned long shift;
 	mpz_ptr pz1, pz2;
- 
+
+	check_exact_args(nargs, "rshift", 2);
+
 	t2 = POP_SCALAR();
 	t1 = POP_SCALAR();
 
@@ -969,7 +1085,7 @@ do_mpfr_rshift(int nargs)
 	/* N.B: See do_mpfp_lshift. */
 	shift = mpz_get_ui(pz2);	/* GMP integer => unsigned long conversion */
 	res = mpg_integer();
-	mpz_fdiv_q_2exp(res->mpg_i, pz1, shift);	/* res = pz1 / 2^shift, round towards −inf */
+	mpz_fdiv_q_2exp(res->mpg_i, pz1, shift);	/* res = pz1 / 2^shift, round towards -inf */
 
 	free_intval(t1, pz1);
 	free_intval(t2, pz2);
@@ -1080,25 +1196,26 @@ do_mpfr_strtonum(int nargs)
 {
 	NODE *tmp, *r;
 
-	tmp = POP_SCALAR();
-	if ((tmp->flags & (NUMBER|NUMCUR)) == 0) {
+	check_exact_args(nargs, "strtonum", 1);
+
+	tmp = fixtype(POP_SCALAR());
+	if ((tmp->flags & NUMBER) == 0) {
 		r = mpg_integer();	/* will be changed to MPFR float if necessary in force_mpnum() */
 		r->stptr = tmp->stptr;
 		r->stlen = tmp->stlen;
 		force_mpnum(r, true, use_lc_numeric);
 		r->stptr = NULL;
 		r->stlen = 0;
+		r->wstptr = NULL;
+		r->wstlen = 0;
+	} else if (is_mpg_float(tmp)) {
+		int tval;
+		r = mpg_float();
+		tval = mpfr_set(r->mpg_numbr, tmp->mpg_numbr, ROUND_MODE);
+		IEEE_FMT(r->mpg_numbr, tval);
 	} else {
-		(void) force_number(tmp);
-		if (is_mpg_float(tmp)) {
-			int tval;
-			r = mpg_float();
-			tval = mpfr_set(r->mpg_numbr, tmp->mpg_numbr, ROUND_MODE);
-			IEEE_FMT(r->mpg_numbr, tval);
-		} else {
-			r = mpg_integer();
-			mpz_set(r->mpg_i, tmp->mpg_i);
-		}
+		r = mpg_integer();
+		mpz_set(r->mpg_i, tmp->mpg_i);
 	}
 
 	DEREF(tmp);
@@ -1118,6 +1235,8 @@ do_mpfr_rand(int nargs ATTRIBUTE_UNUSED)
 	NODE *res;
 	int tval;
 
+	check_exact_args(nargs, "rand", 0);
+
 	if (firstrand) {
 #if 0
 		/* Choose the default algorithm */
@@ -1168,6 +1287,8 @@ do_mpfr_srand(int nargs)
 		firstrand = false;
 	}
 
+	check_args_min_max(nargs, "srand", 0, 1);
+
 	res = mpg_integer();
 	mpz_set(res->mpg_i, seed);	/* previous seed */
 
@@ -1176,7 +1297,7 @@ do_mpfr_srand(int nargs)
 	else {
 		NODE *tmp;
 		tmp = POP_SCALAR();
-		if (do_lint && (tmp->flags & (NUMCUR|NUMBER)) == 0)
+		if (do_lint && (fixtype(tmp)->flags & NUMBER) == 0)
 			lintwarn(_("srand: received non-numeric argument"));
 		force_number(tmp);
 		if (is_mpg_float(tmp))
@@ -1190,6 +1311,100 @@ do_mpfr_srand(int nargs)
 	return res;
 }
 
+#ifdef SUPPLY_INTDIV
+/* do_mpfr_intdiv --- do integer division, return quotient and remainder in dest array */
+
+/*
+ * We define the semantics as:
+ * 	numerator = int(numerator)
+ *	denominator = int(denonmator)
+ *	quotient = int(numerator / denomator)
+ *	remainder = int(numerator % denomator)
+ */
+
+NODE *
+do_mpfr_intdiv(int nargs)
+{
+	NODE *numerator, *denominator, *result;
+	NODE *num, *denom;
+	NODE *quotient, *remainder;
+	NODE *sub, **lhs;
+
+	check_exact_args(nargs, "intdiv", 3);
+
+	result = POP_PARAM();
+	if (result->type != Node_var_array)
+		fatal(_("intdiv: third argument is not an array"));
+	assoc_clear(result);
+
+	denominator = POP_SCALAR();
+	numerator = POP_SCALAR();
+
+	if (do_lint) {
+		if ((fixtype(numerator)->flags & NUMBER) == 0)
+			lintwarn(_("intdiv: received non-numeric first argument"));
+		if ((fixtype(denominator)->flags & NUMBER) == 0)
+			lintwarn(_("intdiv: received non-numeric second argument"));
+	}
+
+	(void) force_number(numerator);
+	(void) force_number(denominator);
+
+	/* convert numerator and denominator to integer */
+	if (is_mpg_integer(numerator)) {
+		num = mpg_integer();
+		mpz_set(num->mpg_i, numerator->mpg_i);
+	} else {
+		if (! mpfr_number_p(numerator->mpg_numbr)) {
+			/* [+-]inf or NaN */
+			unref(numerator);
+			unref(denominator);
+			return make_number((AWKNUM) -1);
+		}
+
+		num = mpg_integer();
+		mpfr_get_z(num->mpg_i, numerator->mpg_numbr, MPFR_RNDZ);
+	}
+
+	if (is_mpg_integer(denominator)) {
+		denom = mpg_integer();
+		mpz_set(denom->mpg_i, denominator->mpg_i);
+	} else {
+		if (! mpfr_number_p(denominator->mpg_numbr)) {
+			/* [+-]inf or NaN */
+			unref(numerator);
+			unref(denominator);
+			unref(num);
+			return make_number((AWKNUM) -1);
+		}
+
+		denom = mpg_integer();
+		mpfr_get_z(denom->mpg_i, denominator->mpg_numbr, MPFR_RNDZ);
+	}
+
+	if (mpz_sgn(denom->mpg_i) == 0)
+		fatal(_("intdiv: division by zero attempted"));
+
+	quotient = mpg_integer();
+	remainder = mpg_integer();
+
+	/* do the division */
+	mpz_tdiv_qr(quotient->mpg_i, remainder->mpg_i, num->mpg_i, denom->mpg_i);
+	unref(num);
+	unref(denom);
+	unref(numerator);
+	unref(denominator);
+
+	sub = make_string("quotient", 8);
+	assoc_set(result, sub, quotient);
+
+	sub = make_string("remainder", 9);
+	assoc_set(result, sub, remainder);
+
+	return make_number((AWKNUM) 0.0);
+}
+#endif /* SUPPLY_INTDIV */
+
 /*
  * mpg_tofloat --- convert an arbitrary-precision integer operand to
  *	a float without loss of precision. It is assumed that the
@@ -1204,13 +1419,13 @@ mpg_tofloat(mpfr_ptr mf, mpz_ptr mz)
 	/*
 	 * When implicitely converting a GMP integer operand to a MPFR float, use
 	 * a precision sufficiently large to hold the converted value exactly.
-	 * 	
+	 *
  	 *	$ ./gawk -M 'BEGIN { print 13 % 2 }'
  	 *	1
  	 * If the user-specified precision is used to convert the integer 13 to a
 	 * float, one will get:
  	 *	$ ./gawk -M 'BEGIN { PREC=2; print 13 % 2.0 }'
- 	 *	0	
+ 	 *	0
  	 */
 
 	prec = mpz_sizeinbase(mz, 2);	/* most significant 1 bit position starting at 1 */
@@ -1224,7 +1439,7 @@ mpg_tofloat(mpfr_ptr mf, mpz_ptr mz)
 	else
 		prec = PRECISION_MIN;
 	/*
-	 * Always set the precision to avoid hysteresis, since do_mpfr_func 
+	 * Always set the precision to avoid hysteresis, since do_mpfr_func
 	 * may copy our precision.
 	 */
 	if (prec != mpfr_get_prec(mf))
@@ -1235,7 +1450,7 @@ mpg_tofloat(mpfr_ptr mf, mpz_ptr mz)
 }
 
 
-/* mpg_add --- add arbitrary-precision numbers */ 
+/* mpg_add --- add arbitrary-precision numbers */
 
 static NODE *
 mpg_add(NODE *t1, NODE *t2)
@@ -1318,7 +1533,7 @@ mpg_mul(NODE *t1, NODE *t2)
 }
 
 
-/* mpg_pow --- exponentiation involving arbitrary-precision numbers */ 
+/* mpg_pow --- exponentiation involving arbitrary-precision numbers */
 
 static NODE *
 mpg_pow(NODE *t1, NODE *t2)
@@ -1370,6 +1585,8 @@ mpg_div(NODE *t1, NODE *t2)
 		mpfr_ptr p1, p2;
 		p1 = MP_FLOAT(t1);
 		p2 = MP_FLOAT(t2);
+		if (mpfr_zero_p(p2))
+			fatal(_("division by zero attempted"));
 		r = mpg_float();
 		tval = mpfr_div(r->mpg_numbr, p1, p2, ROUND_MODE);
 		IEEE_FMT(r->mpg_numbr, tval);
@@ -1403,6 +1620,8 @@ mpg_mod(NODE *t1, NODE *t2)
 		 */
 		NODE *dummy_quotient;
 
+		if (mpz_sgn(t2->mpg_i) == 0)
+			fatal(_("division by zero attempted"));
 		r = mpg_integer();
 		dummy_quotient = mpg_integer();
 		mpz_tdiv_qr(dummy_quotient->mpg_i, r->mpg_i, t1->mpg_i, t2->mpg_i);
@@ -1411,17 +1630,19 @@ mpg_mod(NODE *t1, NODE *t2)
 		mpfr_ptr p1, p2;
 		p1 = MP_FLOAT(t1);
 		p2 = MP_FLOAT(t2);
+		if (mpfr_zero_p(p2))
+			fatal(_("division by zero attempted in `%%'"));
 		r = mpg_float();
 		tval = mpfr_fmod(r->mpg_numbr, p1, p2, ROUND_MODE);
 		IEEE_FMT(r->mpg_numbr, tval);
 	}
 	return r;
 }
-	
+
 /*
  * mpg_interpret --- pre-exec hook in the interpreter. Handles
  *	arithmetic operations with MPFR/GMP numbers.
- */ 
+ */
 
 static int
 mpg_interpret(INSTRUCTION **cp)
@@ -1433,7 +1654,42 @@ mpg_interpret(INSTRUCTION **cp)
 	NODE **lhs;
 	int tval;	/* the ternary value returned by a MPFR function */
 
-	switch ((op = pc->opcode)) {
+	op = pc->opcode;
+	if (do_itrace) {
+		switch (op) {
+		case Op_plus_i:
+		case Op_plus:
+		case Op_minus_i:
+		case Op_minus:
+		case Op_times_i:
+		case Op_times:
+		case Op_exp_i:
+		case Op_exp:
+		case Op_quotient_i:
+		case Op_quotient:
+		case Op_mod_i:
+		case Op_mod:
+		case Op_preincrement:
+		case Op_predecrement:
+		case Op_postincrement:
+		case Op_postdecrement:
+		case Op_unary_minus:
+		case Op_unary_plus:
+		case Op_assign_plus:
+		case Op_assign_minus:
+		case Op_assign_times:
+		case Op_assign_quotient:
+		case Op_assign_mod:
+		case Op_assign_exp:
+			fprintf(stderr, "++ %s: mpg_interpret\n", opcode2str(op));
+			fflush(stderr);
+			break;
+		default:
+			return true;	/* unhandled */
+		}
+	}
+
+	switch (op) {
 	case Op_plus_i:
 		t2 = force_number(pc->memory);
 		goto plus;
@@ -1502,7 +1758,7 @@ quotient:
 		if (op == Op_quotient)
 			DEREF(t2);
 		REPLACE(r);
-		break;		
+		break;
 
 	case Op_mod_i:
 		t2 = force_number(pc->memory);
@@ -1594,9 +1850,32 @@ mod:
 			r = mpg_float();
 			tval = mpfr_neg(r->mpg_numbr, t1->mpg_numbr, ROUND_MODE);
 			IEEE_FMT(r->mpg_numbr, tval);
+		} else {
+			if (! is_zero(t1)) {
+				r = mpg_integer();
+				mpz_neg(r->mpg_i, t1->mpg_i);
+			} else {
+				// have to convert to MPFR for -0.0. sigh
+				r = mpg_float();
+				tval = mpfr_set_d(r->mpg_numbr, 0.0, ROUND_MODE);
+				IEEE_FMT(r->mpg_numbr, tval);
+				tval = mpfr_neg(r->mpg_numbr, r->mpg_numbr, ROUND_MODE);
+				IEEE_FMT(r->mpg_numbr, tval);
+			}
+		}
+		DEREF(t1);
+		REPLACE(r);
+		break;
+
+	case Op_unary_plus:
+		t1 = TOP_NUMBER();
+		if (is_mpg_float(t1)) {
+			r = mpg_float();
+			tval = mpfr_set(r->mpg_numbr, t1->mpg_numbr, ROUND_MODE);
+			IEEE_FMT(r->mpg_numbr, tval);
 		} else {
 			r = mpg_integer();
-			mpz_neg(r->mpg_i, t1->mpg_i);
+			mpz_set(r->mpg_i, t1->mpg_i);
 		}
 		DEREF(t1);
 		REPLACE(r);
@@ -1633,7 +1912,7 @@ mod:
 			r = mpg_pow(t1, t2);
 			break;
 		default:
-			cant_happen();
+			cant_happen("unexpected opcode %s", opcode2str(op));
 		}
 
 		DEREF(t2);
@@ -1684,6 +1963,38 @@ mpfr_unset(NODE *n)
 		mpz_clear(n->mpg_i);
 }
 
+/*
+ * Custom memory allocation functions for GMP / MPFR. We need these so that the
+ * persistent memory feature will also work with the -M option.
+ *
+ * These just call malloc/realloc/free; if we are using PMA then those are
+ * redefined as macros to point at the pma functions, so all should "just work."
+ */
+
+/* mpfr_mem_alloc --- allocate memory */
+
+void *
+mpfr_mem_alloc(size_t alloc_size)
+{
+	return malloc(alloc_size);
+}
+
+/* mpfr_mem_realloc --- reallocate memory */
+
+void *
+mpfr_mem_realloc(void *ptr, size_t old_size, size_t new_size)
+{
+	return realloc(ptr, new_size);
+}
+
+/* mpfr_mem_free --- free memory */
+
+void
+mpfr_mem_free(void *ptr, size_t size)
+{
+	free(ptr);
+}
+
 #else
 
 void
